# Copyright (C) 2010-2015 by Oleksandr Shneyder <oleksandr.shneyder@obviously-nice.de>
# Copyright (C) 2010-2015 by Heinz-M. Graesing <heinz-m.graesing@obviously-nice.de>

# X2Go Admin Server is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# X2Go Admin Server is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# or write to
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.


package x2goadminserver_backend_ldap;

use Exporter 'import';
@EXPORT= qw(getUsers getGroups getGroupsOfUser getGroupsWithUsers modifyUser modifyGroup addUser addGroup removeUsers removeGroups getSessions terminateSession suspendSession);

use strict;
use Config::Simple;
use English qw( -no_match_vars);
use Net::LDAP;
use Crypt::SaltedHash;
use Crypt::SmbHash;
use MIME::Base64;
use Data::Dumper;
use Unicode::Normalize;
use utf8;
use Frontier::RPC2;
use Encode;
use DBI;

my $Config = new Config::Simple(syntax=>'ini');
$Config->read('/etc/x2go/x2goadminserver.conf' ) or die "/etc/x2go/x2goadminserver.conf";
my $url=$Config->param("x2goldap.url");
my $basedn=$Config->param("x2goldap.basedn");
my $sambagroup=$Config->param("x2goldap.smbgroup");
my $sid=$Config->param("x2goldap.sid");

my $dbname=$Config->param("x2goldap.dbname");
my $dbhost=$Config->param("x2goldap.dbhost");
my $dbport=$Config->param("x2goldap.dbport");
my $dbuser=$Config->param("x2goldap.dbuser");
my $dbpass=$Config->param("x2goldap.dbpass");
my $dbssl=$Config->param("x2goldap.dbssl");


my $ldap;

sub terminateSession
{
  my ($login, $password, $server, $id)=@_;
  my $retmsg="OK";

  my ($status,$str)=getServerInfo($login, $password);
  if(!$status)
  {
    return $str;
  }
  system("/etc/x2go/controlsession terminate $server $id");
  return getSessions($login, $password);
}

sub suspendSession
{
  my ($login, $password, $server, $id)=@_;
  my $retmsg="OK";

  my ($status,$str)=getServerInfo($login, $password);
  if(!$status)
  {
    return $str;
  }
  system("/etc/x2go/controlsession suspend $server $id");
  return getSessions($login, $password);
}

sub getSessions
{
  my ($login, $password, @args)=@_;
  my $retmsg="OK";

  my ($status,$str)=getServerInfo($login, $password);
  if(!$status)
  {
    return $str;
  }
  my $dbh=DBI->connect("dbi:Pg:dbname=$dbname;host=$dbhost;port=$dbport;sslmode=$dbssl",
                             "$dbuser", "$dbpass",{AutoCommit => 1}) or return ("DBERR ".$DBI::errstr);

  my @strings;
  my $sth=$dbh->prepare("select session_id, display, server, status,
                        to_char(init_time,'YYYY-MM-DDThh24:MI:SS'), client,
                        to_char(last_time,'YYYY-MM-DDThh24:MI:SS'), uname from sessions");
                        $sth->execute() or return ("DBERR ".$DBI::errstr);
  my @data;
  my $i=0;
  while (@data = $sth->fetchrow_array)
  {
           my ($code, $msg, $cn, $sn)=getUserName("uid=@data[7],ou=People,$basedn");
           if($code)
           {
                  return ("CMDERR "." ".$msg);
           }
           @data[7]=@data[7]."($cn $sn)";
           @strings[$i++]=join('|',@data);
  }
  $sth->finish();
  $dbh->disconnect();
  return join("\n",@strings);
}


sub normalized
{
    my $string=shift;
    my $newstr;
    my %umlaute = ("ä" => "ae", "Ä" => "Ae", "ü" => "ue", "Ü" => "Ue", "ö" => "oe", "Ö" => "Oe", "ß" => "ss" );
    my $umlautkeys = join ("|", keys(%umlaute));
    $string =~ s/($umlautkeys)/$umlaute{$1}/g;
    my @chars = split(//,"$string");
    foreach(@chars)
    {
       my $decomp=NFKD($_);
       $decomp=~ /^(.)/;
       $newstr=$newstr.$1;
    }
    return $newstr;
}


sub getUserName
{
  my $dn=shift;
  my ($cn,$sn);
  my $message=$ldap->search(base => $dn, scope => 'base', filter => '(objectClass=posixAccount)');
  if($message->code)
  {
    return ($message->code,$message->error_desc." ".$dn, $cn, $sn);
  }
  foreach ($message->entries)
  {
    my $asn=$_->{'asn'};
    my $attr=$asn->{'attributes'};
    my @userdata;
    foreach (@$attr)
    {
      my $type=$_->{'type'};
      my $vals=$_->{'vals'};
      if($type eq "givenName")
      {
        $cn=decode("utf-8", @$vals[0]);
      }
      if($type eq "sn")
      {
        $sn=decode("utf-8", @$vals[0]);
      }
    }
  }
  return ($message->code,$message->error_desc, $cn, $sn);
}

sub getGid
{
  my $dn=shift;
  my $gid;
  my $message=$ldap->search(base => $dn, scope => 'base', filter => '(objectClass=posixGroup)');
  if($message->code)
  {
    return ($message->code,$message->error_desc, $gid);
  }
  foreach ($message->entries)
  {
    my $asn=$_->{'asn'};
    my $attr=$asn->{'attributes'};
    my @userdata;
    foreach (@$attr)
    {
      my $type=$_->{'type'};
      my $vals=$_->{'vals'};
      if($type eq "gidNumber")
      {
        $gid=@$vals[0];
      }
    }
  }
  return ($message->code,$message->error_desc, $gid);
}


sub addUserToGroup
{
  my ($dn,$user)=@_;
  my $message=$ldap->modify($dn, add => {memberUid => $user});
  return ($message->code,$message->error_desc);
}

sub removeUserFromGroup
{
  my ($dn,$user)=@_;
  my $message=$ldap->modify($dn, delete => {memberUid => $user});
  return ($message->code,$message->error_desc);
}

sub addUser
{
  return addModifyUser('add',@_);
}

sub modifyUser
{
  return addModifyUser('modify',@_);
}


sub addModifyUser
{

  my ($mode,$login,$password,$user, @args)=@_;
  my $retmsg="OK";

  my ($status,$str)=getServerInfo($login, $password);
  if(!$status)
  {
    return $str;
  }
  my $modsn=0;
  my $setPass=0;
  my $modcn=0;
  my $attr={};
  my @newgrp;
  my @obsgrp;
  my $userUidNumber;
  my $valpwdlastset = 0;
  my $lmPass = "XXX";
  my $ntPass = "XXX";
  my $userPass = "XXX";
  foreach (@args)
  {
    my @cmd=split(":",$_);
    if(@cmd[0] eq "loginshell")
    {
      $attr->{'loginShell'}=@cmd[1];
    }
    if(@cmd[0] eq "uidnumber")
    {
      $attr->{'uidNumber'}=@cmd[1];
      $userUidNumber=@cmd[1];
    }
    if(@cmd[0] eq "homedirectory")
    {
      $attr->{'homeDirectory'}=@cmd[1];
    }
    if(@cmd[0] eq "birthday")
    {
      $attr->{'o'}=@cmd[1];
    }
    if(@cmd[0] eq "password")
    {
      my $csh = Crypt::SaltedHash->new(algorithm => 'SHA-1');
      $csh->add(@cmd[1]);
      $userPass=$csh->generate;
      $attr->{'userPassword'}=$userPass;
      ($lmPass, $ntPass ) = ntlmgen (@cmd[1]);
      $attr->{'sambaNTPassword'}=$ntPass;
      $attr->{'sambaLMPassword'}=$lmPass;
      $valpwdlastset = time;
      $attr->{'sambaPwdLastSet'}=$valpwdlastset;
      $setPass=1;
    }
    if(@cmd[0] eq "fname")
    {
      $attr->{'givenName'}=@cmd[1];
      $modcn=@cmd[1];
    }
    if(@cmd[0] eq "lname")
    {
      $attr->{'sn'}=@cmd[1];
      $modsn=@cmd[1];
    }
    if(@cmd[0] eq "newgroups")
    {
      @newgrp=split(";",@cmd[1]);
    }
    if(@cmd[0] eq "obsoletegroups")
    {
      @obsgrp=split(";",@cmd[1]);
    }
    if(@cmd[0] eq "primarygroup")
    {
      my ($code, $msg, $gid)=getGid("cn=@cmd[1],ou=Group,$basedn");
      if($code)
      {
          return ("CMDERR "." ".$msg);
      }
      $attr->{'gidNumber'}=$gid;
    }
  }
  if($modcn || $modsn)
  {
    if(!($modcn && $modsn))
    {
      my ($code, $msg, $cn, $sn)=getUserName("uid=$user,ou=People,$basedn");
      if($code)
      {
          return ("CMDERR "." ".$msg);
      }
      if(!$modcn)
      {
        $modcn=$cn;
      }
      if(!$modsn)
      {
        $modsn=$sn;
      }
    }
    $attr->{'gecos'}=normalized($modcn." ".$modsn);
    $attr->{'cn'}=$modcn." ".$modsn;
    $attr->{'description'}=$modcn." ".$modsn;
    $attr->{'displayName'}=$modcn." ".$modsn;
  }
  my $message;
  if($mode eq 'modify')
  {
      $message=$ldap->modify("uid=$user,ou=People,$basedn", replace => $attr );
  }
  else
  {
      my $arr;
      while( my ($key, $value) = each (%$attr))
      {
        push(@$arr,$key=>$value);
      }
      push(@$arr,uid=>$user);

      my $winmagic = 2147483647;
      my $valpwdcanchange = 0;
      my $valpwdmustchange = $winmagic;
      my $valpwdlastset = 0;
      my $valacctflags = "[UX]";
      my $userRid=2 * $userUidNumber + 1000;

      push(@$arr, sambaLogonTime => '0');
      push(@$arr, sambaLogoffTime => '2147483647');
      push(@$arr, sambaKickoffTime => '2147483647');
      push(@$arr, sambaPwdCanChange => "$valpwdcanchange");
      push(@$arr, sambaPwdMustChange => "$valpwdmustchange");
      push(@$arr, sambaAcctFlags => "$valacctflags");
      push(@$arr, sambaSID => "$sid-$userRid");
      push(@$arr, sambaPrimaryGroupSID => $sambagroup);

      if(!$setPass)
      {
          push(@$arr, sambaPwdLastSet => "$valpwdlastset");
          push(@$arr, sambaLMPassword => $lmPass);
          push(@$arr, sambaNTPassword => $ntPass);
          push(@$arr, userPassword => $userPass);
      }

      push(@$arr, objectClass=>['top','inetOrgPerson','posixAccount', 'shadowAccount', 'sambaSamAccount']);
      $message=$ldap->add("uid=$user,ou=People,$basedn", attr => $arr);
  }
  if($message->code)
  {
    return ("CMDERR "." ".$message->error_desc());
  }

  if($mode ne 'modify')
  {
      if( -x "/etc/x2go/adduser" )
      {
            system("/etc/x2go/adduser $user $userUidNumber");
      }
  }

  foreach (@newgrp)
  {
      my ($code, $msg)=addUserToGroup("cn=$_,ou=Group,$basedn", $user);
      if($code)
      {
          return ("CMDERR "." ".$msg);
      }
  }

  foreach (@obsgrp)
  {
      my ($code, $msg)=removeUserFromGroup("cn=$_,ou=Group,$basedn", $user);
      if($code)
      {
          return ("CMDERR "." ".$msg);
      }
  }
  $ldap->unbind();
  return $retmsg;
}



sub getGroupsOfUser
{
  my ($login, $password, $user) = @_;
  my ($status,$str)=getServerInfo($login, $password);

  if(!$status)
  {
    return $str;
  }
  my $message=$ldap->search(
    base => "ou=Group,$basedn",
    filter => '(objectClass=posixGroup)'
    );
  $str="";
  if($message->code)
  {
    return ("CMDERR "." ".$message->error_desc());
  }
  foreach ($message->entries)
  {
    my $grp;
    my $exists="0";
    my $gid="0";
    my $asn=$_->{'asn'};
    my $attr=$asn->{'attributes'};
    foreach (@$attr)
    {
      my $type=$_->{'type'};
      my $vals=$_->{'vals'};
      if($type eq "cn")
      {
        $grp=@$vals[0];
      }
      if($type eq "gidNumber")
      {
        $gid=@$vals[0];
      }
      if($type eq "memberUid")
      {
        if(grep { $_ eq $user} @$vals)
        {
          $exists="1";
        }
      }
    }
    $str=$str."\n".$grp.":$exists:$gid";
  }

  $ldap->unbind();
  return $str;
}


sub getGroupsWithUsers
{
  my ($login, $password) = @_;
  my ($status,$str)=getServerInfo($login, $password);
  if(!$status)
  {
    return $str;
  }
  $str=userAttributes();

  $str=$str."\nBEGIN_USERS";
  my ($code, $desc, $result)=getAllUsers($login);
  if($code)
  {
    return ("CMDERR "." ".$desc);
  }
  $str=$str.$result."\nEND_USERS";

  $str=$str."\nBEGIN_GROUPS\n";
  ($code, $desc, $result)=getAllGroupsWithUser($login);
  if($code)
  {
    return ("CMDERR "." ".$desc);
  }
  $str=$str.$result."\nEND_GROUPS\n";
  $ldap->unbind();
  return $str;
}

sub getAllGroupsWithUser
{
  my $login=shift;
  my $str;
  my $message=$ldap->search(
    base => "ou=Group,$basedn",
    filter => '(objectClass=posixGroup)' );

  if(!$message->code)
  {
    foreach ($message->entries)
    {
      my $asn=$_->{'asn'};
      my $attr=$asn->{'attributes'};
      my ($name,$gid,$desc,$members);
      foreach (@$attr)
      {
        my $type=$_->{'type'};
        my $vals=$_->{'vals'};

        if($type eq "cn")
        {
          $name=@$vals[0];
        }
        if($type eq "gidNumber")
        {
          $gid=@$vals[0];
        }
        if($type eq "description")
        {
          $desc=@$vals[0];
        }
        if($type eq "memberUid")
        {
          $members=join(";",@$vals);
        }
      }
      $str=$str."\nNAME=".$name.":GID=".$gid.":DESC=".$desc.":MEMBERS=".$members;
    }
  }
  return ($message->code, $message->error_desc, $str);
}

sub getUsers
{
  my ($login, $password) = @_;
  my ($status,$str)=getServerInfo($login, $password);
  if(!$status)
  {
    return $str;
  }

  $str=userAttributes();
  $str=$str."\nBEGIN_USERS";

  my ($code, $desc, $result)=getAllUsers($login);
  if($code)
  {
    return ("CMDERR "." ".$desc);
  }
  $str=$str.$result."\nEND_USERS";

  $ldap->unbind();

  return $str;
}

sub getAllUsers
{
  my $login=shift;
  my $message=$ldap->search(
    base => "ou=People,$basedn",
    filter => '(objectClass=posixAccount)'   );

  my $str;
  if($message->code)
  {
    return ($message->code, $message->error_desc(),$str);
  }

  foreach ($message->entries)
  {
    my $asn=$_->{'asn'};
    my $attr=$asn->{'attributes'};
    my @userdata;
    foreach (@$attr)
    {
      my $type=$_->{'type'};
      my $vals=$_->{'vals'};
      if($type eq "givenName")
      {
        push(@userdata,decode("utf-8", "FNAME=@$vals[0]"));
      }
      if($type eq "sn")
      {
        push(@userdata,decode("utf-8", "LNAME=@$vals[0]"));
      }
      if($type eq "gidNumber")
      {
        push(@userdata,"GID=@$vals[0]");
      }
      if($type eq "homeDirectory")
      {
        push(@userdata,"HOME=@$vals[0]");
      }
      if($type eq "uid")
      {
        push(@userdata,"UID=@$vals[0]");
      }
      if($type eq "uidNumber")
      {
        push(@userdata,"UIDNUMBER=@$vals[0]");
      }
      if($type eq "loginShell")
      {
        push(@userdata,"SHELL=@$vals[0]");
      }
      if($type eq "o")
      {
        push(@userdata,"BIRTHDAY=@$vals[0]");
      }
    }
    $str=$str."\n".join(":",@userdata);
  }
  return ($message->code, $message->error_desc(),$str);
}

sub getServerInfo
{
  my ($login, $password) = @_;
  $ldap=Net::LDAP->new( $url );
  if(!$ldap)
  {
    return (0, "SERVERDOWN URL:$url ERROR:$@");
  }

  my $message=$ldap->bind( $login, password => $password);
  if($message->code)
  {
    return (0, "NOACCESS"." ".$message->error_desc());
  }
  return (1,"OK");
}


sub userAttributes
{
  return "BEGIN_USERATTR\
uid:edit=0;unicode=0\
uidnumber:edit=0;unicode=0\
firstname:edit=1;unicode=1\
lastname:edit=1;unicode=1\
group:edit=1;unicode=0\
home:edit=1;unicode=0\
shell:edit=1;unicode=0\
birthday:edit=1;unicode=0\
END_USERATTR";
}

sub groupAttributes
{
  return "BEGIN_GROUPATTR\
name:edit=0;unicode=0\
description:edit=1;oncreate=1;unicode=0\
members:edit=1;oncreate=1\
gid:edit=1;oncreate=1\
END_GROUPATTR";
}

sub getGroups
{
  my ($login, $password) = @_;
  my ($status,$str)=getServerInfo($login, $password);
  if(!$status)
  {
    return $str;
  }
  $str=groupAttributes();

  $str=$str."\nBEGIN_USERS";
  my ($code, $desc, $result)=getAllUsers($login);
  if($code)
  {
    return ("CMDERR "." ".$desc);
  }
  $str=$str.$result."\nEND_USERS";

  $str=$str."\nBEGIN_GROUPS\n";
  ($code, $desc, $result)=getAllGroupsWithUser($login);
  if($code)
  {
    return ("CMDERR "." ".$desc);
  }
  $str=$str.$result."\nEND_GROUPS\n";
  $ldap->unbind();
  return $str;
}


sub removeUsers
{
  my ($login, $password, $users) = @_;
  my ($status,$str)=getServerInfo($login, $password);
  if(!$status)
  {
    return $str;
  }
  my @ulist=split(";",$users);
  my $message=$ldap->search(
    base => "ou=Group,$basedn",
    filter => '(objectClass=posixGroup)' );

  if($message->code)
  {
    return ("CMDERR "." ".$message->error_desc());
  }
  $str="OK";
  foreach ($message->entries)
  {
    my $asn=$_->{'asn'};
    my $attr=$asn->{'attributes'};
    foreach (@$attr)
    {
      my $type=$_->{'type'};
      my $vals=$_->{'vals'};
      if($type eq "cn")
      {
        my $grp=@$vals[0];
        my $message=$ldap->modify("cn=$grp,ou=Group,$basedn",
                                  delete => {memberUid => \@ulist});
      }
    }
  }
  foreach(@ulist)
  {
    my $message=$ldap->delete("uid=$_,ou=People,$basedn");
    if($message->code)
    {
       return ("CMDERR "." ".$message->error_desc());
    }
  }
  $ldap->unbind();
  return "$str";
}

sub removeGroups
{
  my ($login, $password, $groups) = @_;
  my ($status,$str)=getServerInfo($login, $password);
  if(!$status)
  {
    return $str;
  }
  my @glist=split(";",$groups);
  foreach(@glist)
  {
    my $message=$ldap->delete("cn=$_,ou=Group,$basedn");
    if($message->code)
    {
       return ("CMDERR "." ".$message->error_desc());
    }
  }
  return "OK";
}


sub modifyGroup
{
  return addModifyGroup('modify',@_);
}

sub addGroup
{
  return addModifyGroup('add',@_);
}

sub addModifyGroup
{
  my ($mode, $login,$password,$group,@args)=@_;
  my ($status,$str)=getServerInfo($login, $password);
  if(!$status)
  {
    return $str;
  }
  my $attr;
  my $addAttr;
  my $rmAttr;
  my $add=0;
  my $rm=0;
  my $mod=0;

  foreach (@args)
  {
      $str=$str."\n".$_;
      my @cmd=split(":",$_);
      if(@cmd[0] eq "description")
      {
        push(@$attr,'description'=>@cmd[1]);
        $mod=1;
      }
      if(@cmd[0] eq "gid")
      {
        push(@$attr,'gidNumber'=>@cmd[1]);
        $mod=1;
      }
      if(@cmd[0] eq "newmembers")
      {
        my $arr;
        @$arr=split(";",@cmd[1]);
        push(@$addAttr,'memberUid'=>$arr);
        $add=1;
      }
      if(@cmd[0] eq "obsoletemembers")
      {
        my $arr;
        @$arr=split(";",@cmd[1]);
        push(@$rmAttr,'memberUid'=>$arr);
        $rm=1;
      }
  }

  my $message;
  if($mode eq 'modify')
  {
      my $changes;
      if($mod)
      {
        push(@$changes,'replace'=>$attr);
      }
      if($add)
      {
        push(@$changes,'add'=>$addAttr);
      }
      if($rm)
      {
        push(@$changes,'delete'=>$rmAttr);
      }
      $message=$ldap->modify("cn=$group,ou=Group,$basedn",
                             changes => $changes );
  }
  else
  {
      push(@$attr,cn=>$group);
      foreach(@$addAttr)
      {
        push(@$attr, $_);
      }
      push(@$attr,objectClass=>['top','posixGroup']);
      $message=$ldap->add("cn=$group,ou=Group,$basedn",
                          attr => $attr);
  }
  if($message->code)
  {
    return ("CMDERR "." ".$message->error_desc());
  }
  $ldap->unbind();
  return $str;
}


return 1;
